@@ -7,12 +7,12 @@ check.is.df <- function(df) {
     return(df)
 }
 
-split.types <- function(df, cols_ord = NULL, cols_ignore = NULL, nthreads = 1) {
+split.types <- function(df, cols_ignore = NULL, nthreads = 1) {
     
     ### validate input data
     df <- check.is.df(df)
     if (NROW(df) < 25)  { stop("Input data has too few row.") }
-    supported_types <- c("numeric", "integer", "character", "factor", "logical", "Date", "POSIXct")
+    supported_types <- c("numeric", "integer", "character", "factor", "ordered", "logical", "Date", "POSIXct")
     coltypes_df     <- Reduce(c, lapply(df, class))
     if (length(setdiff(unique(coltypes_df), c(supported_types, "POSIXt"))) > 0) {
         stop(paste0("Input data can only have the collowing column types: ",
@@ -20,11 +20,6 @@ split.types <- function(df, cols_ord = NULL, cols_ignore = NULL, nthreads = 1) {
                     " - got passed the following: ",
                     paste(setdiff(unique(coltypes_df), supported_types), collapse = ", ")))
     }
-    if (!is.null(cols_ord)) {
-        if (any(!(cols_ord %in% names(df)))) {
-            stop("Ordinal columns passed are not present in data.frame.")
-        }
-    }
     if (!is.null(cols_ignore)) {
         if ("character" %in% class(cols_ignore)) {
             if (length(setdiff(cols_ignore, names(df)))) {
@@ -50,6 +45,7 @@ split.types <- function(df, cols_ord = NULL, cols_ignore = NULL, nthreads = 1) {
     cols_integer <- all_cols[sapply(df, function(x) "integer"   %in% class(x))]
     cols_boolean <- all_cols[sapply(df, function(x) "logical"   %in% class(x))]
     cols_factor  <- all_cols[sapply(df, function(x) "factor"    %in% class(x))]
+    cols_ord     <- all_cols[sapply(df, function(x) "ordered"   %in% class(x))]
     cols_txt     <- all_cols[sapply(df, function(x) "character" %in% class(x))]
     cols_date    <- all_cols[sapply(df, function(x) "Date"      %in% class(x))]
     cols_ts      <- all_cols[sapply(df, function(x) "POSIXct"   %in% class(x))]
@@ -72,11 +68,8 @@ split.types <- function(df, cols_ord = NULL, cols_ignore = NULL, nthreads = 1) {
     if (is.null(outp$cols_date)) { outp$cols_date <- as.character() }
     if (is.null(outp$cols_ts))   { outp$cols_ts   <- as.character() }
     
-    if (!is.null(cols_ord)) {
-        if (!length(cols_factor) || any(!(cols_ord %in% cols_factor))) {
-            stop("Ordinal columns in data.frame must be of type 'factor'.")
-        }
-        outp$cols_cat <- setdiff(outp$cols_cat, cols_ord)
+    if (NROW(outp$cols_cat) && NROW(outp$cols_ord)) {
+        outp$cols_cat <- setdiff(outp$cols_cat, outp$cols_ord)
     }
     
     if (NROW(cols_date)) {
@@ -231,7 +224,7 @@ split.types.new <- function(df, model_data) {
                           !is.na(df[[model_data$cols_ord[[cl]]]])
             df[[model_data$cols_ord[[cl]]]] <- factor(df[[model_data$cols_ord[cl]]], unname(unlist(model_data$ord_levels[[cl]])))
             df[[model_data$cols_ord[[cl]]]] <- ifelse(is.na(df[[model_data$cols_ord[[cl]]]]),
-                                                    -1, as.integer(df[[model_data$cols_ord[[cl]]]]) - 1)
+                                                      -1, as.integer(df[[model_data$cols_ord[[cl]]]]) - 1)
             if (any(new_levels)) {
                 df[[model_data$cols_ord[[cl]]]][new_levels] <- length(model_data$ord_levels[[cl]])
                 throw_new_lev_warn <- TRUE
@@ -385,9 +378,15 @@ report.outliers <- function(lst, rnames, outliers_print) {
                 }
             }
         } else if ("categs_common" %in% names(group_statistics[[row_ix]])) {
-            cat(sprintf("\tdistribution: %.3f%% in [%s]\n",
-                        group_statistics[[row_ix]]$pct_common * 100,
-                        paste(group_statistics[[row_ix]]$categs_common, collapse = ", ")))
+            if (NROW(group_statistics[[row_ix]]$categs_common) == 1) {
+                cat(sprintf("\tdistribution: %.3f%% = [%s]\n",
+                            group_statistics[[row_ix]]$pct_common * 100,
+                            group_statistics[[row_ix]]$categs_common))
+            } else {
+                cat(sprintf("\tdistribution: %.3f%% in [%s]\n",
+                            group_statistics[[row_ix]]$pct_common * 100,
+                            paste(group_statistics[[row_ix]]$categs_common, collapse = ", ")))
+            }
             if (NROW(conditions[[row_ix]])) {
                 cat(sprintf("\t( [norm. obs: %d] - [prior_prob: %.3f%%] - [next smallest: %.3f%%] )\n",
                             group_statistics[[row_ix]]$n_obs,
@@ -398,6 +397,13 @@ report.outliers <- function(lst, rnames, outliers_print) {
                             group_statistics[[row_ix]]$n_obs,
                             group_statistics[[row_ix]]$pct_next_most_comm * 100))
             }
+        }  else if ("categ_maj" %in% names(group_statistics[[row_ix]])) {
+            cat(sprintf("\tdistribution: %.3f%% = [%s]\n",
+                        group_statistics[[row_ix]]$pct_common * 100,
+                        group_statistics[[row_ix]]$categ_maj))
+            cat(sprintf("\t( [norm. obs: %d] - [prior_prob: %.3f%%] )\n",
+                        group_statistics[[row_ix]]$n_obs,
+                        group_statistics[[row_ix]]$prior_prob * 100))
         } else {
             cat(sprintf("\tdistribution: %.3f%% different [norm. obs: %d]",
                         group_statistics[[row_ix]]$pct_other * 100,
@@ -571,12 +577,12 @@ outliers.to.list <- function(df, outliers_info) {
 
 list.to.outliers <- function(outliers_data) {
     return(list(
-        suspicous_value  = sapply(outliers_data, function(x) x$suspicous_value),
-        group_statistics = sapply(outliers_data, function(x) x$group_statistics),
-        conditions       = sapply(outliers_data, function(x) x$conditions),
-        tree_depth       = sapply(outliers_data, function(x) x$tree_depth),
-        uses_NA_branch   = sapply(outliers_data, function(x) x$uses_NA_branch),
-        outlier_score    = sapply(outliers_data, function(x) x$outlier_score)
+        suspicous_value  = sapply(outliers_data, function(x) x$suspicous_value, simplify = FALSE),
+        group_statistics = sapply(outliers_data, function(x) x$group_statistics, simplify = FALSE),
+        conditions       = sapply(outliers_data, function(x) x$conditions, simplify = FALSE),
+        tree_depth       = sapply(outliers_data, function(x) x$tree_depth, simplify = TRUE),
+        uses_NA_branch   = sapply(outliers_data, function(x) x$uses_NA_branch, simplify = TRUE),
+        outlier_score    = sapply(outliers_data, function(x) x$outlier_score, simplify = TRUE)
     ))
 }
 
