@@ -1,12 +1,61 @@
 #' @title Outlier Tree
 #' @description Fit Outlier Tree model to normal data with perhaps some outliers.
-#' @param df  Data Frame with normal data that might contain some outliers.
-#' @param cols_ord Character vector indicating which categorical columns are ordinal.
-#' Ordinal columns must be passed as factors.
+#' @param df  Data Frame with normal data that might contain some outliers. See details for allowed column types.
+#' @param max_depth Maximum depth of the trees to grow. Can also pass zero, in which case it will only look
+#' for outliers with no conditions (i.e. takes each column as a 1-d distribution and looks for outliers in
+#' there independently of the values in other columns).
+#' @param min_gain Minimum gain that a split has to produce in order to consider it (both in terms of looking
+#' for outliers in each branch, and in considering whether to continue branching from them). Note that default
+#' value for GritBot is 1e-6, with `gain_as_pct` = `FALSE`. Recommended to pass higher values (e.g. 1e-1) when using
+#' `gain_as_pct` = `FALSE`.
+#' @param z_norm Maximum Z-value (from standard normal distribution) that can be considered as a normal
+#' observation. Note that simply having values above this will not automatically flag observations as outliers,
+#' nor does it assume that columns follow normal distributions. Also used for categorical and ordinal columns
+#' for building approximate confidence intervals of proportions.
+#' @param z_outlier Minimum Z-value that can be considered as an outlier. There must be a large gap in the
+#' Z-value of the next observation in sorted order to consider it as outlier, given by (z_outlier - z_norm).
+#' Decreasing this parameter is likely to result in more observations being flagged as outliers.
+#' Ignored for categorical and ordinal columns.
+#' @param pct_outliers Approximate max percentage of outliers to expect in a given branch.
+#' @param min_size_numeric Minimum size that branches need to have when splitting a numeric column. In order to look for
+#' outliers in a given branch for a numeric column, it must have a minimum of twice this number
+#' of observations.
+#' @param min_size_categ Minimum size that branches need to have when splitting a categorical or ordinal column. In order to
+#' look for outliers in a given branch for a categorical, ordinal, or boolean column, it must have a minimum of twice
+#' this number of observations.
+#' @param categ_split How to produce categorical-by-categorical splits. Options are:
+#' \itemize{
+#'   \item `"binarize"` : Will binarize the target variable according to whether it's equal to each present category
+#'   within it (greater/less for ordinal), and split each binarized variable separately.
+#'   \item `"bruteforce"` : Will evaluate each possible binary split of the categories (that is, it evaluates 2^n potential
+#'   splits every time). Note that trying this when there are many categories in a column will result
+#'   in exponential computation time that might never finish.
+#'   \item `"separate"` : Will create one branch per category of the splitting variable (this is how GritBot handles them).
+#' }
+#' @param categ_outliers How to look for outliers in categorical variables. Options are:
+#' \itemize{
+#'   \item `"tail"` : Will try to flag outliers if there is a large gap between proportions in sorted order, and this
+#'   gap is unexpected given the prior probabilities. Such criteria tends to sometimes flag too many
+#'   uninteresting outliers, but is able to detect more cases and recognize outliers when there is no
+#'   single dominant category.
+#'   \item `"majority"` : Will calculate an equivalent to z-value according to the number of observations that do not
+#'   belong to the non-majority class, according to formula '(n-n_maj)/(n * p_prior) < 1/z_outlier^2'.
+#'   Such criteria  tends to miss many interesting outliers and will only be able to flag outliers in
+#'   large sample sizes. This is the approach used by GritBot.
+#' }
 #' @param cols_ignore Vector containing columns which will not be split, but will be evaluated for usage
 #' in splitting other columns. Can pass either a logical (boolean) vector with the same number of columns
 #' as `df`, or a character vector of column names (must match with those of `df`).
 #' Pass `NULL` to use all columns.
+#' @param follow_all Whether to continue branching from each split that meets the size and gain criteria.
+#' This will produce exponentially many more branches, and if depth is large, might take forever to finish.
+#' Will also produce a lot more spurious outiers. Not recommended.
+#' @param gain_as_pct Whether the minimum gain above should be taken in absolute terms, or as a percentage of
+#' the standard deviation (for numerical columns) or shannon entropy (for categorical columns). Taking it in
+#' absolute terms will prefer making more splits on columns that have a large variance, while taking it as a
+#' percentage might be more restrictive on them and might create deeper trees in some columns. For GritBot
+#' this parameter would always be `FALSE`. Recommended to pass higher values for `min_gain` when passing `FALSE`
+#' here. Not that when `gain_as_pct` = `FALSE`, the results will be sensitive to the scales of variables.
 #' @param save_outliers Whether to store outliers detected in `df` in the object that is returned.
 #' These outliers can then be extracted from the returned object through function
 #' `extract.training.outliers`.
@@ -14,42 +63,10 @@
 #' the model. Pass zero or `NULL` to avoid printing any. Outliers can be printed from resulting data frame
 #' afterwards through the `predict` method, or through the `print` method (on the extracted outliers, not on
 #' the model object) if passing `save_outliers` = `TRUE`.
-#' @param max_depth Maximum depth of the trees to grow. Can also pass zero, in which case it will only look
-#' for outliers with no conditions (i.e. takes each column as a 1-d distribution and looks for outliers in
-#' there independently of the values in other columns).
-#' @param min_gain Minimum gain that a split has to produce in order to consider it (both in terms of looking
-#' for outliers in each branch, and in considering whether to continue branching from them). Note that default
-#' value for GritBot is 1e-6.
-#' @param z_norm Maximum Z-value (from standard normal distribution) that can be considered as a normal
-#' observation. Note that simply having values above this will not automatically flag observations as outliers,
-#' nor does it assume that columns follow normal distributions. Also used for categorical and ordinal columns
-#' for building approximate confidence intervals of proportions.
-#' @param z_outlier Minimum Z-value that can be considered as an outlier. There must be a large gap in the
-#' Z-value of the next observation in sorted order to consider it as outlier, given by (z_outlier - z_norm).
-#' Ignored for categorical and ordinal columns.
-#' @param pct_outliers Approximate max percentage of outliers to expect in a given branch.
-#' @param min_size_numeric Minimum size that branches need to have when splitting a numeric column.
-#' @param min_size_categ Minimum size that branches need to have when splitting a categorical or ordinal column.
-#' @param categ_as_bin Whether to make categorical-by-categorical binary splits by binarizing each category
-#' in the column and then attempting splits by grouping categories into subsets. Alternative is to create
-#' one branch per category of the column being split from. Ignored when there is only one or fewer categorical
-#' columns. Can only pass one of `categ_as_bin` and `cat_bruteforce_subset`.
-#' @param ord_as_bin Same as `categ_as_bin`, but for ordinal columns, and cumulative (i.e. it splits by "<=",
-#' not "="). Ignored when there are no ordinal columns or no categorical columns.
-#' @param cat_bruteforce_subset Whether to make categorical-by-categorical binary splits by trying all the
-#' possible combinations of columns in each subset (that is, it evaluates 2^n potential splits every time).
-#' Note that trying this when there are many categories in a column will result in exponential computation
-#' time that might never finish. Alternative is to create one branch per category of the column being split
-#' from. Ignored when there is only one or fewer categorical columns. Can only pass one of `categ_as_bin`
-#' and `cat_bruteforce_subset`.
-#' @param follow_all Whether to continue branching from each split that meets the size and gain criteria.
-#' This will produce exponentially many more branches, and if depth is large, might take forever to finish.
-#' Will also produce a lot more spurious outiers. Not recommended.
 #' @param nthreads Number of parallel threads to use. When fitting the model, it will only use up to one
 #' thread per column, while for prediction it will use up to one thread per row. The more threads that are
 #' used, the more memory will be required and allocated, so using more threads will not always lead to better
-#' speed. Passing zero or negative numbers will default to the maximum number of available CPU cores (but not
-#' if the object attribute is overwritten). Can be changed after the object is already initialized.
+#' speed. Can be changed after the object is already initialized.
 #' @return An object with the fitted model that can be used to detect more outliers in new data, and from
 #' which outliers in the training data can be extracted (when passing `save_outliers` = `TRUE`).
 #' @details Explainable outlier detection through decision-tree grouping. Tries to detect outliers by
@@ -61,18 +78,21 @@
 #' Splits are based on gain, while outlierness is based on confidence intervals. Similar in spirit to the GritBot
 #' software developed by RuleQuest research.
 #' 
-#' Supports columns of types numeric, categorical, and ordinal (for this last one, will consider their order when
-#' splitting other columns from them, but not when splitting to "predict" them), and can handle missing values
-#' in any of them. Can also pass dates/timestamps that will get converted to numeric but shown as dates/timestamps
-#' in the output. Offers option to set columns to be used only to split other columns but not to look at outliers
-#' in them.
+#' Supports columns of types numeric (either as type `numeric` or `integer`), categorical (either as type `character` or
+#' `factor` with unordered levels), boolean (as type `logical`), and ordinal (as type `factor` with ordered levels as checked by
+#' `is.ordered`). Can handle missing values in any of them. Can also pass dates/timestamps that will get converted to
+#' numeric but shown as dates/timestamps in the output. Offers option to set columns to be used only for generating
+#' conditions without looking at outliers in them.
 #' 
 #' Infinite values will be taken into consideration when the column is used to split another column
 #' (that is, +inf will go into the branch that is greater than something, -inf into the other branch),
 #' but when a column is the target of the split, they will be taken as missing - that is, it will not report
 #' infinite values as outliers. 
-#' @references GritBot software: \url{https://www.rulequest.com/gritbot-info.html}
-#' @seealso \link{predict.outliertree} \link{extract.training.outliers} \link{hypothyroid}
+#' @references \itemize{
+#'   \item GritBot software: \url{https://www.rulequest.com/gritbot-info.html}
+#'   \item Cortes, David. "Explainable outlier detection through decision tree conditioning." arXiv preprint arXiv:2001.00636 (2020).
+#' }
+#' @seealso \link{predict.outliertree} \link{extract.training.outliers} \link{hypothyroid} \link{unpack.outlier.tree}
 #' @examples 
 #' library(outliertree)
 #' 
@@ -80,7 +100,10 @@
 #' data(hypothyroid)
 #' 
 #' ### fit the model and get a print of outliers
-#' model <- outlier.tree(hypothyroid, outliers_print=10, save_outliers=TRUE)
+#' model <- outlier.tree(hypothyroid,
+#'   outliers_print=10,
+#'   save_outliers=TRUE,
+#'   nthreads=1)
 #' 
 #' ### extract outlier info as R list
 #' outliers <- extract.training.outliers(model)
@@ -95,16 +118,21 @@
 #' outliers.w.names <- predict(model, df.w.names, return_outliers=TRUE)
 #' outliers.w.names[["rownum745"]]
 #' @export
-outlier.tree <- function(df, cols_ord = NULL, cols_ignore = NULL,
+outlier.tree <- function(df, max_depth = 4, min_gain = 1e-2, z_norm = 2.67, z_outlier = 8.0,
+                         pct_outliers = 0.01, min_size_numeric = 25, min_size_categ = 50,
+                         categ_split = "binarize", categ_outliers = "tail", cols_ignore = NULL,
+                         follow_all = FALSE, gain_as_pct = TRUE,
                          save_outliers = FALSE, outliers_print = 10,
-                         max_depth = 4, min_gain = 1e-1, z_norm = 2.67, z_outlier = 8.0,
-                         pct_outliers = 0.01, min_size_numeric = 25, min_size_categ = 75,
-                         categ_as_bin = TRUE, ord_as_bin = TRUE, cat_bruteforce_subset = FALSE,
-                         follow_all = FALSE, nthreads = -1)
+                         nthreads = parallel::detectCores())
 {
     ### validate inputs
-    if ((categ_as_bin | ord_as_bin) & cat_bruteforce_subset) {
-        stop("Can only pass one of 'categ_as_bin/ord_as_bin' and 'cat_bruteforce_subset'.")
+    allowed_cs <- c("binarize", "bruteforce", "separate")
+    if (NROW(categ_split) != 1 || !(categ_split %in% allowed_cs)) {
+        stop(paste0("'categ_split' must be one of ", paste(allowed_cs, collapse = ", ")))
+    }
+    allowed_co <- c("tail", "majority")
+    if (NROW(categ_outliers) != 1 || !(categ_outliers %in% allowed_co)) {
+        stop(paste0("'categ_outliers' must be one of ", paste(allowed_co, collapse = ", ")))
     }
     if (max_depth < 0) { stop("'max_depth' must be >= 0.") }
     if (!("numeric" %in% class(min_gain)))     { stop("'min_gain' must be a decimal number.")     }
@@ -125,26 +153,27 @@ outlier.tree <- function(df, cols_ord = NULL, cols_ignore = NULL,
     if (outliers_print <= 0)     { outliers_print <- FALSE }
     if (outliers_print)          { outliers_print <- as.integer(outliers_print) }
     max_depth <- as.integer(max_depth)
-    min_gain <- as.numeric(min_gain)
-    z_norm <- as.numeric(z_norm)
+    min_gain  <- as.numeric(min_gain)
+    z_norm    <- as.numeric(z_norm)
     z_outlier <- as.numeric(z_outlier)
-    pct_outliers <- as.numeric(pct_outliers)
+    pct_outliers     <- as.numeric(pct_outliers)
     min_size_numeric <- as.integer(min_size_numeric)
-    min_size_categ <- as.integer(min_size_categ)
-    categ_as_bin <- as.logical(categ_as_bin)
-    ord_as_bin <- as.logical(ord_as_bin)
-    cat_bruteforce_subset <- as.logical(cat_bruteforce_subset)
-    follow_all <- as.logical(follow_all)
-    
+    min_size_categ   <- as.integer(min_size_categ)
+    categ_split      <- categ_split
+    categ_outliers   <- categ_outliers
+    follow_all  <- as.logical(follow_all)
+    gain_as_pct <- as.logical(gain_as_pct)
+
     ### decompose data into C arrays and names, then pass to C++
-    model_data <- split.types(df, cols_ord, cols_ignore, nthreads)
+    model_data <- split.types(df, cols_ignore, nthreads)
     model_data$obj_from_cpp <- fit_OutlierTree(model_data$arr_num, model_data$ncol_num,
                                                model_data$arr_cat, model_data$ncol_cat, model_data$ncat,
                                                model_data$arr_ord, model_data$ncol_ord, model_data$ncat_ord,
                                                model_data$nrow, model_data$cols_ign, nthreads,
-                                               categ_as_bin, ord_as_bin, cat_bruteforce_subset,
+                                               categ_split == "binarize", categ_split == "binarize",
+                                               categ_split == "bruteforce", categ_outliers == "majority",
                                                max_depth, pct_outliers, min_size_numeric, min_size_categ,
-                                               min_gain, follow_all, z_norm, z_outlier,
+                                               min_gain, follow_all, gain_as_pct, z_norm, z_outlier,
                                                as.logical(save_outliers | outliers_print),
                                                lapply(model_data$cat_levels, as.character),
                                                lapply(model_data$ord_levels, as.character),
@@ -199,7 +228,7 @@ outlier.tree <- function(df, cols_ord = NULL, cols_ignore = NULL,
 #' @details Note that after loading a serialized object from `outlier.tree` through `readRDS` or `load`,
 #' it will only de-serialize the underlying C++ object upon running `predict` or `print`, so the first run will
 #' be slower, while subsequent runs will be faster as the C++ object will already be in-memory.
-#' @seealso \link{outlier.tree}
+#' @seealso \link{outlier.tree} \link{print.outlieroutputs} \link{unpack.outlier.tree}
 #' @examples 
 #' library(outliertree)
 #' ### random data frame with an obvious outlier
@@ -222,7 +251,7 @@ outlier.tree <- function(df, cols_ord = NULL, cols_ignore = NULL,
 #' )
 #'     
 #' ### fit model on training data
-#' outliers_model = outlier.tree(df, outliers_print = FALSE)
+#' outliers_model = outlier.tree(df, outliers_print=FALSE, nthreads=1)
 #' 
 #' ### find the test outlier
 #' test_outliers = predict(outliers_model, df_test,
@@ -231,7 +260,7 @@ outlier.tree <- function(df, cols_ord = NULL, cols_ignore = NULL,
 #' ### retrieve the outlier info (for row 1) as an R list
 #' test_outliers[[1]]
 #' @export 
-predict.outliertree <- function(object, newdata, outliers_print = 15, return_outliers = FALSE, ...) {
+predict.outliertree <- function(object, newdata, outliers_print = 15, return_outliers = TRUE, ...) {
     check.is.model.obj(object)
     if (check_null_ptr_model(object$obj_from_cpp$ptr_model)) {
         ptr_new <- deserialize_OutlierTree(object$obj_from_cpp$serialized_obj)
@@ -244,7 +273,9 @@ predict.outliertree <- function(object, newdata, outliers_print = 15, return_out
         if (outliers_print) {
             report.no.outliers()
         }
-        if (return_outliers) { return(produce.empty.outliers(row.names(newdata))) } else { return(invisible(NULL)) }
+        if (return_outliers) {
+            return(invisible(produce.empty.outliers(row.names(newdata))))
+        } else { return(invisible(NULL)) }
     }
     
     c_arr_data    <- split.types.new(newdata, object)
@@ -266,7 +297,7 @@ predict.outliertree <- function(object, newdata, outliers_print = 15, return_out
     }
     if (return_outliers) {
         outliers_info$found_outliers <- NULL
-        return(outliers.to.list(newdata, outliers_info))
+        return(invisible(outliers.to.list(newdata, outliers_info)))
     }
 }
 
@@ -280,6 +311,27 @@ predict.outliertree <- function(object, newdata, outliers_print = 15, return_out
 #' data frame were null, or the row names they had if non-null). Pass `NULL` to print information
 #' about potentially all rows
 #' @param ... Not used.
+#' @return No return value.
+#' @seealso \link{outlier.tree} \link{predict.outliertree}
+#' @examples 
+#' ### Example re-printing results for selected rows
+#' library(outliertree)
+#' data("hypothyroid")
+#' 
+#' ### Fit model
+#' otree <- outlier.tree(hypothyroid,
+#'   nthreads=1,
+#'   outliers_print=0)
+#'   
+#' ### Store predictions
+#' pred <- predict(otree,
+#'   hypothyroid,
+#'   outliers_print=0,
+#'   return_outliers=TRUE)
+#'   
+#' ### Print stored predictions
+#' ### Row 531 is an outlier, but 532 is not
+#' print(pred, only_these_rows = c(531, 532))
 #' @export 
 print.outlieroutputs <- function(x, outliers_print = 15, only_these_rows = NULL, ...) {
     if (NROW(x) == 0) { report.no.outliers(); return(invisible(NULL)); }
@@ -322,3 +374,64 @@ check.outlierness.bounds <- function(outlier_tree_model) {
     check.is.model.obj(outlier_tree_model)
     return(outlier_tree_model$obj_from_cpp$bounds)
 }
+
+#' @title Unpack Outlier Tree model after de-serializing
+#' @description  After persisting an outlier tree model object through `saveRDS`, `save`, or restarting a session, the
+#' underlying C++ objects that constitute the outlier tree model and which live only on the C++ heap memory are not saved along,
+#' thus not restored after loading a saved model through `readRDS` or `load`.
+#' 
+#' The model object however keeps serialized versions of the C++ objects as raw bytes, from which the C++ objects can be
+#' reconstructed, and are done so automatically after calling `predict`, `print`, or `summary` on the
+#' freshly-loaded object from `readRDS` or `load`.
+#' 
+#' But due to R's environments system (as opposed to other systems such as Python which can use pass-by-reference), they will
+#' only be re-constructed in the environment that is calling `predict`, `print`, etc. and not in higher-up environments
+#' (i.e. if you call `predict` on the object from inside different functions, each function will have to reconstruct the
+#' C++ objects independently and they will only live within the function that called `predict`).
+#' 
+#' This function serves as an environment-level unpacker that will reconstruct the C++ object in the environment in which
+#' it is called (i.e. if you need to call `predict` from inside multiple functions, use this function before passing the
+#' freshly-loaded model object to those other functions, and then they will not need to reconstruct the C++ objects anymore),
+#' in the same way as `predict` or `print`, but without producing any outputs or messages.
+#' @param model An Outlier Tree object as returned by `outlier.tree`, which has been just loaded from a disk
+#' file through `readRDS`, `load`, or a session restart.
+#' @return No return value. Object is modified in-place.
+#' @examples 
+#' ### Warning: this example will generate a temporary .Rds
+#' ### file in your temp folder, and will then delete it
+#' library(outliertree)
+#' set.seed(1)
+#' df <- as.data.frame(matrix(rnorm(1000), nrow = 250))
+#' otree <- outlier.tree(df, outliers_print=0, nthreads=1)
+#' temp_file <- file.path(tempdir(), "otree.Rds")
+#' saveRDS(otree, temp_file)
+#' otree2 <- readRDS(temp_file)
+#' file.remove(temp_file)
+#' 
+#' ### will de-serialize inside, but object is short-lived
+#' wrap_predict <- function(model, data) {
+#'     pred <- predict(model, data, outliers_print = 0)
+#'     cat("pointer inside function is this: ")
+#'     print(model$obj_from_cpp$ptr_model)
+#'     return(pred)
+#' }
+#' temp <- wrap_predict(otree2, df)
+#' cat("pointer outside function is this: \n")
+#' print(otree2$obj_from_cpp$ptr_model) ### pointer to the C++ object
+#' 
+#' ### now unpack the C++ object beforehand
+#' unpack.outlier.tree(otree2)
+#' print("after unpacking beforehand")
+#' temp <- wrap_predict(otree2, df)
+#' cat("pointer outside function is this: \n")
+#' print(otree2$obj_from_cpp$ptr_model)
+#' @export
+unpack.outlier.tree <- function(model)  {
+    check.is.model.obj(model)
+    if (check_null_ptr_model(model$obj_from_cpp$ptr_model)) {
+        ptr_new <- deserialize_OutlierTree(model$obj_from_cpp$serialized_obj)
+        eval.parent(substitute(model$obj_from_cpp$ptr_model <- ptr_new))
+        model$obj_from_cpp$ptr_model <- ptr_new
+    }
+    return(invisible(NULL))
+}
