@@ -49,6 +49,16 @@ Rcpp::LogicalVector check_null_ptr_model(SEXP ptr_model)
     return Rcpp::LogicalVector(R_ExternalPtrAddr(ptr_model) == NULL);
 }
 
+double* set_R_nan_as_C_nan(double *restrict x_R, std::vector<double> &x_C, size_t n, int nthreads)
+{
+    x_C.assign(x_R, x_R + n);
+    #pragma omp parallel for schedule(static) num_threads(nthreads) shared(x_R, x_C, n)
+    for (size_t_for i = 0; i < n; i++)
+        if (isnan(x_R[i]))
+            x_C[i] = NAN;
+    return x_C.data();
+}
+
 
 /* for predicting outliers */
 Rcpp::List describe_outliers(ModelOutputs &model_outputs,
@@ -67,7 +77,7 @@ Rcpp::List describe_outliers(ModelOutputs &model_outputs,
     size_t ncols_num     = model_outputs.ncols_numeric;
     size_t ncols_cat     = model_outputs.ncols_categ;
     size_t ncols_num_num = model_outputs.ncols_numeric - min_date.size() - min_ts.size();
-    size_t ncols_date    = model_outputs.ncols_numeric - ncols_num_num - min_ts.size();
+    size_t ncols_date    = min_date.size();
     size_t ncols_cat_cat = cat_levels.size();
     Rcpp::List outp;
     
@@ -159,14 +169,16 @@ Rcpp::List describe_outliers(ModelOutputs &model_outputs,
                 } else if (outl_col < (ncols_num_num + ncols_date)) {
                     if (arr_num[row + outl_col * nrows] >= model_outputs.all_clusters[outl_col][outl_clust].upper_lim) {
                         lst_stats[row] = Rcpp::List::create(
-                            Rcpp::_["upper_thr"] = Rcpp::Date(model_outputs.all_clusters[outl_col][outl_clust].display_lim_high - 1 + min_date[outl_col - ncols_num_num]),
+                            Rcpp::_["upper_thr"] = Rcpp::Date(model_outputs.all_clusters[outl_col][outl_clust].display_lim_high
+                                                              - 1 + min_date[outl_col - ncols_num_num]),
                             Rcpp::_["pct_below"] = Rcpp::wrap(model_outputs.all_clusters[outl_col][outl_clust].perc_below),
                             Rcpp::_["mean"]      = Rcpp::Date(model_outputs.all_clusters[outl_col][outl_clust].display_mean - 1 + min_date[outl_col - ncols_num_num]),
                             Rcpp::_["n_obs"]     = Rcpp::wrap((int)model_outputs.all_clusters[outl_col][outl_clust].cluster_size)
                         );
                     } else {
                         lst_stats[row] = Rcpp::List::create(
-                            Rcpp::_["lower_thr"] = Rcpp::Date(model_outputs.all_clusters[outl_col][outl_clust].display_lim_low - 1 + min_date[outl_col - ncols_num_num]),
+                            Rcpp::_["lower_thr"] = Rcpp::Date(model_outputs.all_clusters[outl_col][outl_clust].display_lim_low
+                                                              - 1 + min_date[outl_col - ncols_num_num]),
                             Rcpp::_["pct_above"] = Rcpp::wrap(model_outputs.all_clusters[outl_col][outl_clust].perc_above),
                             Rcpp::_["mean"]      = Rcpp::Date(model_outputs.all_clusters[outl_col][outl_clust].display_mean - 1 + min_date[outl_col - ncols_num_num]),
                             Rcpp::_["n_obs"]     = Rcpp::wrap((int)model_outputs.all_clusters[outl_col][outl_clust].cluster_size)
@@ -175,16 +187,20 @@ Rcpp::List describe_outliers(ModelOutputs &model_outputs,
                 } else {
                     if (arr_num[row + outl_col * nrows] >= model_outputs.all_clusters[outl_col][outl_clust].upper_lim) {
                         lst_stats[row] = Rcpp::List::create(
-                            Rcpp::_["upper_thr"] = Rcpp::Datetime(model_outputs.all_clusters[outl_col][outl_clust].display_lim_high - 1 + min_ts[outl_col - ncols_num_num - ncols_date]),
+                            Rcpp::_["upper_thr"] = Rcpp::Datetime(model_outputs.all_clusters[outl_col][outl_clust].display_lim_high
+                                                                  - 1 + min_ts[outl_col - ncols_num_num - ncols_date]),
                             Rcpp::_["pct_below"] = Rcpp::wrap(model_outputs.all_clusters[outl_col][outl_clust].perc_below),
-                            Rcpp::_["mean"]      = Rcpp::Datetime(model_outputs.all_clusters[outl_col][outl_clust].display_mean - 1 + min_ts[outl_col - ncols_num_num - ncols_date]),
+                            Rcpp::_["mean"]      = Rcpp::Datetime(model_outputs.all_clusters[outl_col][outl_clust].display_mean
+                                                                  - 1 + min_ts[outl_col - ncols_num_num - ncols_date]),
                             Rcpp::_["n_obs"]     = Rcpp::wrap((int)model_outputs.all_clusters[outl_col][outl_clust].cluster_size)
                         );
                     } else {
                         lst_stats[row] = Rcpp::List::create(
-                            Rcpp::_["lower_thr"] = Rcpp::Datetime(model_outputs.all_clusters[outl_col][outl_clust].display_lim_low - 1 + min_ts[outl_col - ncols_num_num - ncols_date]),
+                            Rcpp::_["lower_thr"] = Rcpp::Datetime(model_outputs.all_clusters[outl_col][outl_clust].display_lim_low
+                                                                  - 1 + min_ts[outl_col - ncols_num_num - ncols_date]),
                             Rcpp::_["pct_above"] = Rcpp::wrap(model_outputs.all_clusters[outl_col][outl_clust].perc_above),
-                            Rcpp::_["mean"]      = Rcpp::Datetime(model_outputs.all_clusters[outl_col][outl_clust].display_mean - 1 + min_ts[outl_col - ncols_num_num - ncols_date]),
+                            Rcpp::_["mean"]      = Rcpp::Datetime(model_outputs.all_clusters[outl_col][outl_clust].display_mean
+                                                                  - 1 + min_ts[outl_col - ncols_num_num - ncols_date]),
                             Rcpp::_["n_obs"]     = Rcpp::wrap((int)model_outputs.all_clusters[outl_col][outl_clust].cluster_size)
                         );
                     }
@@ -198,14 +214,26 @@ Rcpp::List describe_outliers(ModelOutputs &model_outputs,
                             }
                         }
                     if (model_outputs.all_clusters[outl_col][outl_clust].split_type != Root) {
-                        lst_stats[row] = Rcpp::List::create(
-                            Rcpp::_["categs_common"]      = Rcpp::as<Rcpp::CharacterVector>(cat_levels[outl_col - ncols_num][tmp_bool]),
-                            Rcpp::_["pct_common"]         = Rcpp::wrap(model_outputs.all_clusters[outl_col][outl_clust].perc_in_subset),
-                            Rcpp::_["pct_next_most_comm"] = Rcpp::wrap(model_outputs.all_clusters[outl_col][outl_clust].perc_next_most_comm),
-                            Rcpp::_["prior_prob"]         = Rcpp::wrap(model_outputs.prop_categ[model_outputs.start_ix_cat_counts[outl_col - ncols_num] +
-                                                                       arr_cat[row + (outl_col - ncols_num) * nrows]]),
-                            Rcpp::_["n_obs"]              = Rcpp::wrap(model_outputs.all_clusters[outl_col][outl_clust].cluster_size)
-                        );
+                        if (model_outputs.all_clusters[outl_col][outl_clust].categ_maj < 0) {
+                            lst_stats[row] = Rcpp::List::create(
+                                Rcpp::_["categs_common"]      = Rcpp::as<Rcpp::CharacterVector>(cat_levels[outl_col - ncols_num][tmp_bool]),
+                                Rcpp::_["pct_common"]         = Rcpp::wrap(model_outputs.all_clusters[outl_col][outl_clust].perc_in_subset),
+                                Rcpp::_["pct_next_most_comm"] = Rcpp::wrap(model_outputs.all_clusters[outl_col][outl_clust].perc_next_most_comm),
+                                Rcpp::_["prior_prob"]         = Rcpp::wrap(model_outputs.prop_categ[model_outputs.start_ix_cat_counts[outl_col - ncols_num] +
+                                                                           arr_cat[row + (outl_col - ncols_num) * nrows]]),
+                                Rcpp::_["n_obs"]              = Rcpp::wrap(model_outputs.all_clusters[outl_col][outl_clust].cluster_size)
+                            );
+                        } else {
+                            lst_stats[row] = Rcpp::List::create(
+                                Rcpp::_["categ_maj"]   = Rcpp::as<Rcpp::CharacterVector>(cat_levels[outl_col - ncols_num][
+                                                                                                model_outputs.all_clusters[outl_col][outl_clust].categ_maj
+                                                                                                ]),
+                                Rcpp::_["pct_common"]  = Rcpp::wrap(model_outputs.all_clusters[outl_col][outl_clust].perc_in_subset),
+                                Rcpp::_["prior_prob"]  = Rcpp::wrap(model_outputs.prop_categ[model_outputs.start_ix_cat_counts[outl_col - ncols_num] +
+                                                                    arr_cat[row + (outl_col - ncols_num) * nrows]]),
+                                Rcpp::_["n_obs"]       = Rcpp::wrap(model_outputs.all_clusters[outl_col][outl_clust].cluster_size)
+                            );
+                        }
                     } else {
                         lst_stats[row] = Rcpp::List::create(
                             Rcpp::_["categs_common"]      = Rcpp::as<Rcpp::CharacterVector>(cat_levels[outl_col - ncols_num][tmp_bool]),
@@ -230,14 +258,26 @@ Rcpp::List describe_outliers(ModelOutputs &model_outputs,
                     }
                 }
                 if (model_outputs.all_clusters[outl_col][outl_clust].split_type != Root) {
-                    lst_stats[row] = Rcpp::List::create(
-                        Rcpp::_["categs_common"]      = Rcpp::as<Rcpp::CharacterVector>(ord_levels[outl_col - ncols_num - ncols_cat][tmp_bool]),
-                        Rcpp::_["pct_common"]         = Rcpp::wrap(model_outputs.all_clusters[outl_col][outl_clust].perc_in_subset),
-                        Rcpp::_["pct_next_most_comm"] = Rcpp::wrap(model_outputs.all_clusters[outl_col][outl_clust].perc_next_most_comm),
-                        Rcpp::_["prior_prob"]         = Rcpp::wrap(model_outputs.prop_categ[model_outputs.start_ix_cat_counts[outl_col - ncols_num] +
-                                                                   arr_ord[row + (outl_col - ncols_num - ncols_cat) * nrows]]),
-                        Rcpp::_["n_obs"]              = Rcpp::wrap(model_outputs.all_clusters[outl_col][outl_clust].cluster_size)
-                    );
+                    if (model_outputs.all_clusters[outl_col][outl_clust].categ_maj < 0) {
+                        lst_stats[row] = Rcpp::List::create(
+                            Rcpp::_["categs_common"]      = Rcpp::as<Rcpp::CharacterVector>(ord_levels[outl_col - ncols_num - ncols_cat][tmp_bool]),
+                            Rcpp::_["pct_common"]         = Rcpp::wrap(model_outputs.all_clusters[outl_col][outl_clust].perc_in_subset),
+                            Rcpp::_["pct_next_most_comm"] = Rcpp::wrap(model_outputs.all_clusters[outl_col][outl_clust].perc_next_most_comm),
+                            Rcpp::_["prior_prob"]         = Rcpp::wrap(model_outputs.prop_categ[model_outputs.start_ix_cat_counts[outl_col - ncols_num] +
+                                                                       arr_ord[row + (outl_col - ncols_num - ncols_cat) * nrows]]),
+                            Rcpp::_["n_obs"]              = Rcpp::wrap(model_outputs.all_clusters[outl_col][outl_clust].cluster_size)
+                        );
+                    } else {
+                        lst_stats[row] = Rcpp::List::create(
+                            Rcpp::_["categ_maj"]   = Rcpp::as<Rcpp::CharacterVector>(ord_levels[outl_col - ncols_num - ncols_cat][
+                                                                                        model_outputs.all_clusters[outl_col][outl_clust].categ_maj
+                                                                                        ]),
+                            Rcpp::_["pct_common"]  = Rcpp::wrap(model_outputs.all_clusters[outl_col][outl_clust].perc_in_subset),
+                            Rcpp::_["prior_prob"]  = Rcpp::wrap(model_outputs.prop_categ[model_outputs.start_ix_cat_counts[outl_col - ncols_num] +
+                                                                arr_ord[row + (outl_col - ncols_num - ncols_cat) * nrows]]),
+                            Rcpp::_["n_obs"]       = Rcpp::wrap(model_outputs.all_clusters[outl_col][outl_clust].cluster_size)
+                        );
+                    }
                 } else {
                     lst_stats[row] = Rcpp::List::create(
                         Rcpp::_["categs_common"]      = Rcpp::as<Rcpp::CharacterVector>(ord_levels[outl_col - ncols_num - ncols_cat][tmp_bool]),
@@ -281,7 +321,12 @@ Rcpp::List describe_outliers(ModelOutputs &model_outputs,
                                 cond_clust["value_this"] = Rcpp::as<Rcpp::CharacterVector>(NA_STRING);
                             }
                         } else {
-                            cond_clust["value_this"] = Rcpp::wrap((bool)arr_cat[row + model_outputs.all_clusters[outl_col][outl_clust].col_num * nrows]);
+
+                            if (arr_cat[row + model_outputs.all_clusters[outl_col][outl_clust].col_num * nrows] >= 0) {
+                                cond_clust["value_this"] = Rcpp::wrap((bool)arr_cat[row + model_outputs.all_clusters[outl_col][outl_clust].col_num * nrows]);
+                            } else {
+                                cond_clust["value_this"] = Rcpp::LogicalVector(1, NA_LOGICAL);
+                            }
                         }
                         break;
                     }
@@ -317,9 +362,9 @@ Rcpp::List describe_outliers(ModelOutputs &model_outputs,
                             case Categorical:
                             {
                                 if (model_outputs.all_clusters[outl_col][outl_clust].col_num < ncols_cat_cat) {
-                                    cond_clust["value_comp"] = Rcpp::wrap(NA_REAL);
+                                    cond_clust["value_comp"] = Rcpp::wrap(NA_STRING);
                                 } else {
-                                    cond_clust["value_comp"] = Rcpp::wrap(NA_LOGICAL);
+                                    cond_clust["value_comp"] = Rcpp::LogicalVector(1, NA_LOGICAL);
                                 }
                                 break;
                             }
@@ -416,7 +461,7 @@ Rcpp::List describe_outliers(ModelOutputs &model_outputs,
                                                                                               [model_outputs.all_clusters[outl_col][outl_clust].split_lev]);
                             } else {
                                 cond_clust["comparison"] = Rcpp::CharacterVector("=");
-                                cond_clust["value_comp"] = Rcpp::wrap((bool)model_outputs.all_clusters[outl_col][outl_clust].split_lev);
+                                cond_clust["value_comp"] = Rcpp::wrap((bool) model_outputs.all_clusters[outl_col][outl_clust].split_lev);
                             }
                         } else {
                             cond_clust["comparison"] = Rcpp::CharacterVector("=");
@@ -556,9 +601,9 @@ Rcpp::List describe_outliers(ModelOutputs &model_outputs,
                                             cond_clust["comparison"] = Rcpp::CharacterVector("is NA");
                                             cond_clust["value_comp"] = Rcpp::as<Rcpp::CharacterVector>(NA_STRING);
                                         } else {
-                                            cond_clust["value_this"] = Rcpp::wrap(NA_LOGICAL);
+                                            cond_clust["value_this"] = Rcpp::LogicalVector(1, NA_LOGICAL);
                                             cond_clust["comparison"] = Rcpp::CharacterVector("is NA");
-                                            cond_clust["value_comp"] = Rcpp::wrap(NA_LOGICAL);
+                                            cond_clust["value_comp"] = Rcpp::LogicalVector(1, NA_LOGICAL);
                                         }
                                         break;
                                     }
@@ -754,9 +799,9 @@ Rcpp::List describe_outliers(ModelOutputs &model_outputs,
                                             cond_clust["comparison"] = Rcpp::CharacterVector("is NA");
                                             cond_clust["value_comp"] = Rcpp::as<Rcpp::CharacterVector>(NA_STRING);
                                         } else {
-                                            cond_clust["value_this"] = Rcpp::wrap(NA_LOGICAL);
+                                            cond_clust["value_this"] = Rcpp::LogicalVector(1, NA_LOGICAL);
                                             cond_clust["comparison"] = Rcpp::CharacterVector("is NA");
-                                            cond_clust["value_comp"] = Rcpp::wrap(NA_LOGICAL);
+                                            cond_clust["value_comp"] = Rcpp::LogicalVector(1, NA_LOGICAL);
                                         }
                                         break;
                                     }
@@ -854,7 +899,7 @@ Rcpp::List describe_outliers(ModelOutputs &model_outputs,
                                 } else {
                                     cond_clust["value_this"] = Rcpp::wrap((bool) arr_cat[row + model_outputs.all_trees[outl_col][parent_tree].col_num * nrows]);
                                     cond_clust["comparison"] = Rcpp::CharacterVector("=");
-                                    cond_clust["value_comp"] = Rcpp::wrap(model_outputs.all_trees[outl_col][parent_tree].split_subset[1]);
+                                    cond_clust["value_comp"] = Rcpp::wrap((bool) model_outputs.all_trees[outl_col][parent_tree].split_subset[1]);
                                 }
                                 break;
                             }
@@ -875,7 +920,7 @@ Rcpp::List describe_outliers(ModelOutputs &model_outputs,
                                 } else {
                                     cond_clust["value_this"] = Rcpp::wrap((bool) arr_cat[row + model_outputs.all_trees[outl_col][parent_tree].col_num * nrows]);
                                     cond_clust["comparison"] = Rcpp::CharacterVector("=");
-                                    cond_clust["value_comp"] = Rcpp::wrap(model_outputs.all_trees[outl_col][parent_tree].split_subset[0]);
+                                    cond_clust["value_comp"] = Rcpp::wrap((bool) model_outputs.all_trees[outl_col][parent_tree].split_subset[0]);
                                 }
                                 break;
                             }
@@ -892,7 +937,7 @@ Rcpp::List describe_outliers(ModelOutputs &model_outputs,
                                     } else {
                                         cond_clust["value_this"] = Rcpp::wrap((bool) arr_cat[row + model_outputs.all_trees[outl_col][parent_tree].col_num * nrows]);
                                         cond_clust["comparison"] = Rcpp::CharacterVector("=");
-                                        cond_clust["value_comp"] = Rcpp::wrap(model_outputs.all_trees[outl_col][parent_tree].split_subset[1]);
+                                        cond_clust["value_comp"] = Rcpp::wrap((bool) model_outputs.all_trees[outl_col][parent_tree].split_subset[1]);
                                     }
                                 } else {
                                     cond_clust["value_this"] = Rcpp::CharacterVector(1, ord_levels[model_outputs.all_trees[outl_col][parent_tree].col_num]
@@ -916,7 +961,7 @@ Rcpp::List describe_outliers(ModelOutputs &model_outputs,
                                     } else {
                                         cond_clust["value_this"] = Rcpp::wrap((bool) arr_cat[row + model_outputs.all_trees[outl_col][parent_tree].col_num * nrows]);
                                         cond_clust["comparison"] = Rcpp::CharacterVector("=");
-                                        cond_clust["value_comp"] = Rcpp::wrap(model_outputs.all_trees[outl_col][parent_tree].split_subset[0]);
+                                        cond_clust["value_comp"] = Rcpp::wrap((bool) model_outputs.all_trees[outl_col][parent_tree].split_subset[0]);
                                     }
                                 } else {
                                     cond_clust["value_this"] = Rcpp::CharacterVector(1, ord_levels[model_outputs.all_trees[outl_col][parent_tree].col_num]
@@ -1042,9 +1087,9 @@ Rcpp::List fit_OutlierTree(Rcpp::NumericVector arr_num, size_t ncols_numeric,
                            Rcpp::IntegerVector arr_cat, size_t ncols_categ,   Rcpp::IntegerVector ncat,
                            Rcpp::IntegerVector arr_ord, size_t ncols_ord,     Rcpp::IntegerVector ncat_ord,
                            size_t nrows, Rcpp::LogicalVector cols_ignore_r, int nthreads,
-                           bool categ_as_bin, bool ord_as_bin, bool cat_bruteforce_subset,
+                           bool categ_as_bin, bool ord_as_bin, bool cat_bruteforce_subset, bool categ_from_maj,
                            size_t max_depth, double max_perc_outliers, size_t min_size_numeric, size_t min_size_categ,
-                           double min_gain, bool follow_all, double z_norm, double z_outlier,
+                           double min_gain, bool follow_all, bool gain_as_pct, double z_norm, double z_outlier,
                            bool return_outliers,
                            Rcpp::ListOf<Rcpp::StringVector> cat_levels,
                            Rcpp::ListOf<Rcpp::StringVector> ord_levels,
@@ -1064,15 +1109,18 @@ Rcpp::List fit_OutlierTree(Rcpp::NumericVector arr_num, size_t ncols_numeric,
         for (size_t cl = 0; cl < tot_cols; cl++) cols_ignore[cl] = (bool) cols_ignore_r[cl];
         cols_ignore_ptr = &cols_ignore[0];
     }
+    std::vector<double> Xcpp;
+    double *arr_num_C = set_R_nan_as_C_nan(&arr_num[0], Xcpp, arr_num.size(), nthreads);
+
     std::unique_ptr<ModelOutputs> model_outputs = std::unique_ptr<ModelOutputs>(new ModelOutputs);
     found_outliers = fit_outliers_models(*model_outputs,
-                                         &arr_num[0], ncols_numeric,
+                                         arr_num_C, ncols_numeric,
                                          &arr_cat[0], ncols_categ, &ncat[0],
                                          &arr_ord[0], ncols_ord,   &ncat_ord[0],
                                          nrows, cols_ignore_ptr, nthreads,
-                                         categ_as_bin, ord_as_bin, cat_bruteforce_subset,
+                                         categ_as_bin, ord_as_bin, cat_bruteforce_subset, categ_from_maj,
                                          max_depth, max_perc_outliers, min_size_numeric, min_size_categ,
-                                         min_gain, follow_all, z_norm, z_outlier);
+                                         min_gain, gain_as_pct, follow_all, z_norm, z_outlier);
 
     outp["bounds"] = extract_outl_bounds(*model_outputs,
                                          cat_levels,
@@ -1083,7 +1131,7 @@ Rcpp::List fit_OutlierTree(Rcpp::NumericVector arr_num, size_t ncols_numeric,
     outp["serialized_obj"] = serialize_OutlierTree(model_outputs.get());
     if (return_outliers) {
         outp["outliers_info"] = describe_outliers(*model_outputs,
-                                                  &arr_num[0],
+                                                  arr_num_C,
                                                   &arr_cat[0],
                                                   &arr_ord[0],
                                                   cat_levels,
@@ -1120,11 +1168,14 @@ Rcpp::List predict_OutlierTree(SEXP ptr_model, size_t nrows, int nthreads,
                                Rcpp::NumericVector min_date,
                                Rcpp::NumericVector min_ts)
 {
+    std::vector<double> Xcpp;
+    double *arr_num_C = set_R_nan_as_C_nan(&arr_num[0], Xcpp, arr_num.size(), nthreads);
+
     ModelOutputs *model_outputs = static_cast<ModelOutputs*>(R_ExternalPtrAddr(ptr_model));
     bool found_outliers = find_new_outliers(&arr_num[0], &arr_cat[0], &arr_ord[0],
                                             nrows, nthreads, *model_outputs);
     Rcpp::List outp = describe_outliers(*model_outputs,
-                                        &arr_num[0],
+                                        arr_num_C,
                                         &arr_cat[0],
                                         &arr_ord[0],
                                         cat_levels,
